# Standard Library Imports
from urllib.parse import parse_qsl
from urllib.parse import urlencode
from urllib.parse import urlparse
from urllib.parse import urlunparse

# Third Party Imports
from fastapi.dependencies.utils import request_params_to_args
from fastapi.encoders import jsonable_encoder
from fastapi.routing import APIRoute

# Imports from this repository
from fastapi_cloud_tasks.exception import MissingParamError
from fastapi_cloud_tasks.exception import WrongTypeError
from fastapi_cloud_tasks.utils import err_val

try:
    # Third Party Imports
    import ujson as json
except Exception:
    # Standard Library Imports
    import json


class Requester:
    def __init__(
        self,
        *,
        route: APIRoute,
        base_url: str,
    ) -> None:
        self.route = route
        self.base_url = base_url.rstrip("/")

    def _headers(self, *, values):
        headers = err_val(request_params_to_args(self.route.dependant.header_params, values))
        cookies = err_val(request_params_to_args(self.route.dependant.cookie_params, values))
        if len(cookies) > 0:
            headers["Cookies"] = "; ".join([f"{k}={v}" for (k, v) in cookies.items()])
        # We use json only.
        headers["Content-Type"] = "application/json"
        # Always send string headers and skip all headers which are supposed to be sent by cloudtasks
        return {str(k): str(v) for (k, v) in headers.items() if not str(k).startswith("x_cloudtasks_")}

    def _url(self, *, values):
        route = self.route
        path_values = err_val(request_params_to_args(route.dependant.path_params, values))
        for (name, converter) in route.param_convertors.items():
            if name in path_values:
                continue
            if name not in values:
                raise MissingParamError(param=name)

            # TODO: should we catch errors here and raise better errors?
            path_values[name] = converter.convert(values[name])
        path = route.path_format.format(**path_values)
        params = err_val(request_params_to_args(route.dependant.query_params, values))

        # Make final URL

        # Split base url into parts
        url_parts = list(urlparse(self.base_url))

        # Add relative path
        # Note: you might think urljoin is a better solution here, it is not.
        url_parts[2] = url_parts[2].strip("/") + "/" + path.strip("/")

        # Make query dict and update our with our params
        query = dict(parse_qsl(url_parts[4]))
        query.update(params)

        # override query params
        url_parts[4] = urlencode(query)
        return urlunparse(url_parts)

    def _body(self, *, values):
        body = None
        body_field = self.route.body_field
        if body_field and body_field.name:
            got_body = values.get(body_field.name, None)
            if got_body is None:
                if body_field.required:
                    raise MissingParamError(name=body_field.name)
                got_body = body_field.get_default()
            if not isinstance(got_body, body_field.type_):
                raise WrongTypeError(field=body_field.name, type=body_field.type_)
            body = json.dumps(jsonable_encoder(got_body)).encode()
        return body



# Standard Library Imports

# Third Party Imports
from fastapi.routing import APIRoute
from google.cloud import scheduler_v1
from google.protobuf import duration_pb2

# Imports from this repository
from fastapi_cloud_tasks.hooks import ScheduledHook
from fastapi_cloud_tasks.requester import Requester
from fastapi_cloud_tasks.utils import schedulerMethod


class Scheduler(Requester):
    def __init__(
        self,
        *,
        route: APIRoute,
        base_url: str,
        location_path: str,
        schedule: str,
        client: scheduler_v1.CloudSchedulerClient,
        pre_create_hook: ScheduledHook,
        name: str = "",
        job_create_timeout: float = 10.0,
        retry_config: scheduler_v1.RetryConfig = None,
        time_zone: str = "UTC",
        force: bool = False,
    ) -> None:
        super().__init__(route=route, base_url=base_url)
        if name == "":
            name = route.unique_id

        if retry_config is None:
            retry_config = scheduler_v1.RetryConfig(
                retry_count=5,
                max_retry_duration=duration_pb2.Duration(seconds=0),
                min_backoff_duration=duration_pb2.Duration(seconds=5),
                max_backoff_duration=duration_pb2.Duration(seconds=120),
                max_doublings=5,
            )

        self.retry_config = retry_config
        location_parts = client.parse_common_location_path(location_path)

        self.job_id = client.job_path(job=name, **location_parts)
        self.time_zone = time_zone

        self.location_path = location_path
        self.cron_schedule = schedule
        self.job_create_timeout = job_create_timeout

        self.method = schedulerMethod(route.methods)
        self.client = client
        self.pre_create_hook = pre_create_hook
        self.force = force

    def schedule(self, **kwargs):
        # Create http request
        request = scheduler_v1.HttpTarget()
        request.http_method = self.method
        request.uri = self._url(values=kwargs)
        request.headers = self._headers(values=kwargs)

        body = self._body(values=kwargs)
        if body:
            request.body = body

        # Scheduled the task
        job = scheduler_v1.Job(
            name=self.job_id,
            http_target=request,
            schedule=self.cron_schedule,
            retry_config=self.retry_config,
            time_zone=self.time_zone,
        )

        request = scheduler_v1.CreateJobRequest(parent=self.location_path, job=job)

        request = self.pre_create_hook(request)

        if self.force or self._has_changed(request=request):
            # Delete and create job
            self.delete()
            self.client.create_job(request=request, timeout=self.job_create_timeout)

    def _has_changed(self, request: scheduler_v1.CreateJobRequest):
        try:
            job = self.client.get_job(name=request.job.name)
            # Remove things that are either output only or GCP adds by default
            job.user_update_time = None
            job.state = None
            job.status = None
            job.last_attempt_time = None
            job.schedule_time = None
            del job.http_target.headers["User-Agent"]
            # Proto compare works directly with `__eq__`
            return job != request.job
        except Exception:
            return True
        return False

    def delete(self):
        # We return true or exception because you could have the delete code on multiple instances
        try:
            self.client.delete_job(name=self.job_id, timeout=self.job_create_timeout)
            return True
        except Exception as ex:
            return ex



# Standard Library Imports
from typing import Callable

# Third Party Imports
from fastapi.routing import APIRoute
from google.cloud import scheduler_v1

# Imports from this repository
from fastapi_cloud_tasks.hooks import ScheduledHook
from fastapi_cloud_tasks.hooks import noop_hook
from fastapi_cloud_tasks.scheduler import Scheduler


def ScheduledRouteBuilder(
    *,
    base_url: str,
    location_path: str,
    job_create_timeout: float = 10.0,
    pre_create_hook: ScheduledHook = None,
    client=None,
):

    if client is None:
        client = scheduler_v1.CloudSchedulerClient()

    if pre_create_hook is None:
        pre_create_hook = noop_hook

    class ScheduledRouteMixin(APIRoute):
        def get_route_handler(self) -> Callable:
            original_route_handler = super().get_route_handler()
            self.endpoint.scheduler = self.schedulerOptions
            return original_route_handler

        def schedulerOptions(self, *, name, schedule, **options) -> Scheduler:
            schedulerOpts = dict(
                base_url=base_url,
                location_path=location_path,
                client=client,
                pre_create_hook=pre_create_hook,
                job_create_timeout=job_create_timeout,
                name=name,
                schedule=schedule,
            )

            schedulerOpts.update(options)

            return Scheduler(route=self, **schedulerOpts)

    return ScheduledRouteMixin



# Standard Library Imports
import datetime

# Third Party Imports
from fastapi.routing import APIRoute
from google.cloud import tasks_v2
from google.protobuf import timestamp_pb2

# Imports from this repository
from fastapi_cloud_tasks.hooks import DelayedTaskHook
from fastapi_cloud_tasks.requester import Requester
from fastapi_cloud_tasks.utils import taskMethod


class Delayer(Requester):
    def __init__(
        self,
        *,
        route: APIRoute,
        base_url: str,
        queue_path: str,
        task_create_timeout: float = 10.0,
        client: tasks_v2.CloudTasksClient,
        pre_create_hook: DelayedTaskHook,
        countdown: int = 0,
        task_id: str = None,
    ) -> None:
        super().__init__(route=route, base_url=base_url)
        self.queue_path = queue_path
        self.countdown = countdown
        self.task_create_timeout = task_create_timeout

        self.task_id = task_id
        self.method = taskMethod(route.methods)
        self.client = client
        self.pre_create_hook = pre_create_hook

    def delay(self, **kwargs):
        # Create http request
        request = tasks_v2.HttpRequest()
        request.http_method = self.method
        request.url = self._url(values=kwargs)
        request.headers = self._headers(values=kwargs)

        body = self._body(values=kwargs)
        if body:
            request.body = body

        # Scheduled the task
        task = tasks_v2.Task(http_request=request)
        schedule_time = self._schedule()
        if schedule_time:
            task.schedule_time = schedule_time

        # Make task name for deduplication
        if self.task_id:
            task.name = f"{self.queue_path}/tasks/{self.task_id}"

        request = tasks_v2.CreateTaskRequest(parent=self.queue_path, task=task)

        request = self.pre_create_hook(request)

        return self.client.create_task(request=request, timeout=self.task_create_timeout)

    def _schedule(self):
        if self.countdown is None or self.countdown <= 0:
            return None
        d = datetime.datetime.utcnow() + datetime.timedelta(seconds=self.countdown)
        timestamp = timestamp_pb2.Timestamp()
        timestamp.FromDatetime(d)
        return timestamp



# Standard Library Imports
from typing import Dict
from typing import List
from typing import Tuple

# Third Party Imports
from google.cloud import scheduler_v1
from google.cloud import tasks_v2
from pydantic.error_wrappers import ErrorWrapper

# Imports from this repository
from fastapi_cloud_tasks.exception import BadMethodException


def location_path(*, project, location):
    return scheduler_v1.CloudSchedulerClient.common_location_path(project=project, location=location)


def queue_path(*, project, location, queue):
    return tasks_v2.CloudTasksClient.queue_path(project=project, location=location, queue=queue)


def err_val(resp: Tuple[Dict, List[ErrorWrapper]]):
    values, errors = resp

    if len(errors) != 0:
        # TODO: Log everything but raise first only
        # TODO: find a better way to raise and display these errors
        raise errors[0].exc
    return values


def taskMethod(methods):
    methodMap = {
        "POST": tasks_v2.HttpMethod.POST,
        "GET": tasks_v2.HttpMethod.GET,
        "HEAD": tasks_v2.HttpMethod.HEAD,
        "PUT": tasks_v2.HttpMethod.PUT,
        "DELETE": tasks_v2.HttpMethod.DELETE,
        "PATCH": tasks_v2.HttpMethod.PATCH,
        "OPTIONS": tasks_v2.HttpMethod.OPTIONS,
    }
    methods = list(methods)
    # Only crash if we're being bound
    if len(methods) > 1:
        raise BadMethodException("Can't trigger task with multiple methods")
    method = methodMap.get(methods[0], None)
    if method is None:
        raise BadMethodException(f"Unknown method {methods[0]}")
    return method


def schedulerMethod(methods):
    methodMap = {
        "POST": scheduler_v1.HttpMethod.POST,
        "GET": scheduler_v1.HttpMethod.GET,
        "HEAD": scheduler_v1.HttpMethod.HEAD,
        "PUT": scheduler_v1.HttpMethod.PUT,
        "DELETE": scheduler_v1.HttpMethod.DELETE,
        "PATCH": scheduler_v1.HttpMethod.PATCH,
        "OPTIONS": scheduler_v1.HttpMethod.OPTIONS,
    }
    methods = list(methods)
    # Only crash if we're being bound
    if len(methods) > 1:
        raise BadMethodException("Can't schedule task with multiple methods")
    method = methodMap.get(methods[0], None)
    if method is None:
        raise BadMethodException(f"Unknown method {methods[0]}")
    return method



# Standard Library Imports
from typing import Callable

# Third Party Imports
from google.cloud import scheduler_v1
from google.cloud import tasks_v2
from google.protobuf import duration_pb2

DelayedTaskHook = Callable[[tasks_v2.CreateTaskRequest], tasks_v2.CreateTaskRequest]
ScheduledHook = Callable[[scheduler_v1.CreateJobRequest], scheduler_v1.CreateJobRequest]


def noop_hook(request):
    return request


def chained_hook(*hooks):
    def chain(request):
        for hook in hooks:
            request = hook(request)
        return request

    return chain


def oidc_scheduled_hook(token: scheduler_v1.OidcToken) -> ScheduledHook:
    def add_token(
        request: scheduler_v1.CreateJobRequest,
    ) -> scheduler_v1.CreateJobRequest:
        request.job.http_target.oidc_token = token
        return request

    return add_token


def oidc_delayed_hook(token: tasks_v2.OidcToken) -> DelayedTaskHook:
    def add_token(request: tasks_v2.CreateTaskRequest) -> tasks_v2.CreateTaskRequest:
        request.task.http_request.oidc_token = token
        return request

    return add_token


def oauth_scheduled_hook(token: scheduler_v1.OAuthToken) -> ScheduledHook:
    def add_token(
        request: scheduler_v1.CreateJobRequest,
    ) -> scheduler_v1.CreateJobRequest:
        request.job.http_target.oauth_token = token
        return request

    return add_token


def oauth_delayed_hook(token: tasks_v2.OAuthToken) -> DelayedTaskHook:
    def add_token(request: tasks_v2.CreateTaskRequest) -> tasks_v2.CreateTaskRequest:
        request.task.http_request.oauth_token = token
        return request

    return add_token


def deadline_scheduled_hook(duration: duration_pb2.Duration) -> ScheduledHook:
    def deadline(
        request: scheduler_v1.CreateJobRequest,
    ) -> scheduler_v1.CreateJobRequest:
        request.job.attempt_deadline = duration
        return request

    return deadline


def deadline_delayed_hook(duration: duration_pb2.Duration) -> DelayedTaskHook:
    def deadline(request: tasks_v2.CreateTaskRequest) -> tasks_v2.CreateTaskRequest:
        request.task.dispatch_deadline = duration
        return request

    return deadline



# Standard Library Imports
from typing import Callable

# Third Party Imports
from fastapi.routing import APIRoute
from google.cloud import scheduler_v1
from google.cloud import tasks_v2

# Imports from this repository
from fastapi_cloud_tasks.delayer import Delayer
from fastapi_cloud_tasks.hooks import DelayedTaskHook
from fastapi_cloud_tasks.hooks import noop_hook
from fastapi_cloud_tasks.scheduler import Scheduler


def DelayedRouteBuilder(
    *,
    base_url: str,
    queue_path: str,
    task_create_timeout: float = 10.0,
    pre_create_hook: DelayedTaskHook = None,
    client=None,
    scheduler_client=None,
):
    if client is None:
        client = tasks_v2.CloudTasksClient()

    if pre_create_hook is None:
        pre_create_hook = noop_hook

    class TaskRouteMixin(APIRoute):
        def get_route_handler(self) -> Callable:
            original_route_handler = super().get_route_handler()
            self.endpoint.options = self.delayOptions
            self.endpoint.delay = self.delay
            return original_route_handler

        def delayOptions(self, **options) -> Delayer:
            delayOpts = dict(
                base_url=base_url,
                queue_path=queue_path,
                task_create_timeout=task_create_timeout,
                client=client,
                pre_create_hook=pre_create_hook,
            )
            if hasattr(self.endpoint, "_delayOptions"):
                delayOpts.update(self.endpoint._delayOptions)
            delayOpts.update(options)

            return Delayer(
                route=self,
                **delayOpts,
            )

        def delay(self, **kwargs):
            return self.delayOptions().delay(**kwargs)

    return TaskRouteMixin



# Third Party Imports
from setuptools import setup

with open("version.txt") as f:
    version = f.read().strip()

with open("README.md", encoding="utf-8") as f:
    long_description = f.read().strip()


setup(
    name="fastapi-cloud-tasks",
    version=version,
    description="Trigger delayed Cloud Tasks from FastAPI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    licesnse="MIT",
    packages=["fastapi_cloud_tasks"],
    install_requires=["google-cloud-tasks", "google-cloud-scheduler", "fastapi"],
    test_requires=[],
    zip_safe=False,
)



# Third Party Imports
from pydantic.errors import MissingError
from pydantic.errors import PydanticValueError


class MissingParamError(MissingError):
    msg_template = "field required: {param}"


class WrongTypeError(PydanticValueError):
    msg_template = "Expected {field} to be of type {type}"


class BadMethodException(Exception):
    pass
